#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

## Copyright 2002-2003 by LivingLogic AG, Bayreuth/Germany.
## Copyright 2002-2003 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVINGLOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


"""
<par>This module implements <app>make</app> functionality.</par>

<par>Like <app>make</app> it allows you to specify dependencies
between files and actions to be executed when files don't exist or are out of date
with respect to one of their sources. But unlike <app>make</app> you
can do this in a object oriented way and targets are not only limited to files,
but you can implement e.g. dependencies on database records by implementing new
<pyref class="Target"><class>Target</class></pyref> classes.</par>

<par>Relevant classes are:
<ulist>
<item><pyref class="Project"><class>Project</class></pyref>, which is the container for all targets,
actions and dependencies in a project,</item>
<item><pyref class="Target"><class>Target</class></pyref> (and subclasses), which are files
(or other entities like database records) that have to be built by actions,</item>
<item><pyref class="Dep"><class>Dep</class></pyref> (and subclasses) which specify types
of dependencies between targets and</item>
<item><pyref class="Action"><class>Action</class></pyref> (and subclasses) which specify how targets
are (re)built if they don't exist or aren't up to date or how they should be modified.</item>
</ulist>
</par>
"""


import sys, os, getopt, warnings, re, datetime

from ll import ansistyle, url


def dtnow():
	"""
	Return the current time (in &utc;).
	"""
	return datetime.datetime.utcnow()


indent = " "*8


###
### exceptions & warnings
###

class MakeWarning(Warning):
	"""
	Base class for all warnings in <app>make</app>.
	"""


class RedefinedTargetWarning(MakeWarning):
	"""
	Warning that will be issued when a target is added to a project and a target
	with the same id already exists
	"""

	def __init__(self, targetid):
		self.targetid = targetid
	def __str__(self):
		return "target with id=%r redefined" % self.targetid


class ClockSkewWarning(MakeWarning):
	"""
	Warning that will be issued when a target has a last modification date from the future.
	"""

	def __init__(self, targetid, lastmodified, now):
		self.targetid = targetid
		self.lastmodified = lastmodified
		self.now = now
	def __str__(self):
		return "target %r is from the future: %r>%r=now" % (self.targetid, self.lastmodified, self.now)


class UndefinedTargetError(KeyError):
	"""
	Exception that will be raised, when a target with the specified id doesn't exist within the project.
	"""

	def __init__(self, id):
		self.id = id

	def __str__(self):
		return "target %r undefined" % self.id


###
### Dependency classes
###

class Dep(object):
	"""
	<par>A <class>Dep</class> is a dependency between a <z>source</z>
	and a <z>target</z> <pyref class="Target"><class>Target</class></pyref>. It's possible
	to assign specific meanings to certain dependencies by using subclasses of
	<class>Dep</class>.</par>
	"""
	__slots__ = ("source", "target")

	def __init__(self, source, target):
		"""
		<par>create a <class>Dep</class> instance.</par>
		"""
		self.source = source
		self.target = target
		target._sources.append(self)
		source._targets.append(self)

	def __hash__(self):
		return hash(self.source.id) ^ hash(self.target.id)

	def __eq__(self, other):
		return self.source.id == other.source.id and self.target.id == other.target.id

	def __ne__(self, other):
		return self.source.id != other.source.id or self.target.id != other.target.id

	def __repr__(self):
		return "<%s from %r to %r at 0x%x>" % (self.__class__.__name__, self.source, self.target, id(self))


class MainDep(Dep):
	"""
	<par>This class can be used to create a special dependency between
	two targets. This is e.g. used by <pyref class="CopyAction"><class>CopyAction</class></pyref>
	and <pyref class="XISTAction"><class>XISTAction</class></pyref>.</par>
	"""
	pass


class OrderedDep(MainDep):
	"""
	<par>This dependency class represents an ordered dependency, i.e. each
	instance has an <lit>order</lit> attribute.</par>
	"""
	order = 0

	__slots__ = ("order")

	def __init__(self, source, target, order=None):
		"""
		<par>create a new <class>OrderedDep</class> instance.
		When <arg>order</arg> is not specified a new unique order value
		will be used that will be greater than the order value of
		previously created <class>OrderedDep</class>
		instances.</par>
		"""
		MainDep.__init__(self, source, target)
		if order is None:
			order = self.__class__.order
			self.__class__.order += 1
		self.order = order


###
### Target classes
###

class Target(object):
	def __init__(self, project, id, actions=None):
		"""
		<par>Create a new <class>Target</class> instance.</par>
		<par><arg>actions</arg> can be either:</par>
		<ulist>
		<item><lit>None</lit> in which case this target can not be
		rebuilt, but must exist and be up to date. In most cases
		these targets don't have any sources anyway.</item>
		<item>One <pyref class="Action"><class>Action</class></pyref>,
		which will be <pyref class="Action" method="execute"><method>execute</method>d</pyref>
		to rebuild the target.</item>
		<item>A sequence of <class>Action</class> objects, in which
		case all actions will be executed sequentially. E.g. the first
		action might create the target file and subsequent actions can
		then modify this file (e.g. changing the access permissions
		(via <pyref class="ModeAction"><class>ModeAction</class></pyref>)
		or the owner (via <pyref class="OwnerAction"><class>OwnerAction</class></pyref>).</item>
		</ulist>
		"""
		self.project = project
		self.id = id
		if isinstance(actions, Action):
			actions = [actions]
		elif actions is None:
			actions = []
		self.actions = actions
		self._sources = []
		self._targets = []
		self.project[self.id] = self

	def lastmodified(self):
		"""
		<par>return the timestamp, when this target was last modified or <lit>None</lit>
		when the target does not exist. The timestamp will be cached and will be updated
		when a target is remade.</par>
		"""
		try:
			return self._lastmodified
		except AttributeError:
			lm = self._fetchlastmodified()
			now = dtnow()
			if lm is not None and lm>now:
				self.project.reportWarning(ClockSkewWarning(self, lm, now), 2)
			self._lastmodified = lm
			return lm

	def _fetchlastmodified(self):
		"""
		<par>return the timestamp of the last modification.
		Overwrite in subclasses.</par>
		"""
		raise NotImplementedError()

	def touch(self):
		"""
		<par>set the <z>last modified</z> timestamp to now.</par>
		"""
		self._lastmodified = dtnow()

	def dirty(self):
		"""
		<par>forget the cached <z>last modifified</z> timestamp,
		so the target might be rebuilt on subsequent calls.</par>
		"""
		try:
			del self._lastmodified
		except AttributeError:
			pass

	def clear(self):
		"""
		<par>Called by <pyref class="Project" method="clear"><method>Project.clear</method></pyref>
		to get rid of all the old targets.</par>
		"""
		self._sources = []
		self._targets = []

	def update(self):
		"""
		<par>Update this target and all of its sources, if necessary.</par>
		"""
		t1 = dtnow()
		for dep in self._sources:
			dep.source.update()
		if self.hasToBeRebuilt():
			if not self.actions:
				raise ValueError("don't know how to build %s" % self)
			if self.project.showTargets:
				msg = str(self)
				if self.project.showActions:
					msg += " <-%s--" % "--".join([str(self.project.actionStr(action)) for action in  self.actions])
				elif self.project.showSources:
					msg += " <--"
				if self.project.showSources:
					if self.lastmodified() is None:
						reason = "(target doesn't exist)"
					else:
						newer = map(str, self.newersources())
						if len(newer):
							if len(newer)>self.project.maxNewerReport:
								reason = "(%d sources changed)" % len(newer)
							else:
								reason = ", ".join(newer)
						else:
							reason = ""
					msg += " %s" % reason
				msg += " ...\n"
				self.project.report(msg)
			failed = False
			v = []
			try:
				for action in self.actions:
					msg = action.execute(self)
					if msg is not None:
						v.append(msg)
			except SystemExit:
				raise
			except KeyboardInterrupt:
				raise
			except Exception, ex:
				if self.project.ignoreExceptions:
					failed = True
				else:
					raise
			t2 = dtnow()
			msg = "; ".join(v)
			if failed:
				import traceback
				traceback.print_exc()
				fullmsg = "%s%s after %s\n" % (indent, self.project.errorStr("failed"), self.project.secondsStr(t2-t1))
				self.project.reportError(fullmsg)
				self.project.targetsfailed += 1
			else:
				if self.project.showMessages:
					fullmsg = "%sdone after %s" % (indent, self.project.secondsStr(t2-t1))
					if msg:
						fullmsg += ": %s" % msg
					fullmsg += "\n"
					self.project.report(fullmsg)
				self.project.targetsrebuilt += 1
			self.touch() # always touch the target, so building can continue even if this target failed

	def itersources(self, type=Dep):
		"""
		<par>Iterate through all sources of <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._sources:
			if isinstance(dep, type):
				yield dep.source

	def sources(self, type=Dep):
		"""
		<par>return a list of all the sources for <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.itersources(type))

	def itersourcedeps(self, type=Dep):
		"""
		<par>Iterate through all the source dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._sources:
			if isinstance(dep, type):
				yield dep

	def sourcedeps(self, type=Dep):
		"""
		<par>return a list of all the source dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.itersourcedeps(type))

	def itertargetdeps(self, type=Dep):
		"""
		<par>Iterate through all the target dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._targets:
			if isinstance(dep, type):
				yield dep

	def targetdeps(self, type=Dep):
		"""
		<par>return a list of all the targets for <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.itertargetdeps(type))

	def itertargets(self, type=Dep):
		"""
		<par>Iterate through all targets of <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._targets:
			if isinstance(dep, type):
				yield dep.target

	def targets(self, type=Dep):
		"""
		<par>return a list of all the target dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.itertargets(type))

	def newersourcedeps(self, type=Dep):
		"""
		<par>return a list of all the sources that or newer the this one
		(or don't exist).</par>
		"""
		selfLM = self.lastmodified()
		sources = self.sourcedeps(type)
		if selfLM is None:
			return sources
		else:
			newer = []
			for dep in sources:
				sourceLM = dep.source.lastmodified()
				if  sourceLM is None or sourceLM>selfLM:
					newer.append(dep)
			return newer

	def newersources(self, type=Dep):
		"""
		<par>return a list of all the sources that or newer the this one
		(or don't exist).</par>
		"""
		return [ dep.source for dep in self.newersourcedeps(type) ]

	def dependOn(self, *others):
		"""
		<par>add one or more dependencies.</par>

		<par><arg>others</arg> may contain <pyref class="Target"><class>Target</class></pyref>
		instances and subclasses of <pyref class="Dep"><class>Dep</class></pyref>. This will be used to
		create specific dependencies. Initially <pyref class="Dep"><class>Dep</class></pyref> will be used to
		construct source dependencies for every <pyref class="Target"><class>Target</class></pyref> instance
		encountered in the argument list. From the point in the argument list where a
		<pyref class="Dep"><class>Dep</class></pyref> subclass is encountered this new class will be
		used for constructing dependencies. This makes it possible to switch between different
		dependency classes.</par>
		"""
		dependencyClass = Dep
		for other in others:
			if isinstance(other, type) and issubclass(other, Dep):
				dependencyClass = other
			else:
				dependency = dependencyClass(other, self)

	def hasToBeRebuilt(self):
		"""
		<par>does <self/> have to be rebuilt (either because this target does not exist, or
		it is older than one of its sources)?</par>
		"""
		selfLM = self.lastmodified()
		if selfLM is not None:
			return len(self.newersourcedeps())>0
		else:
			return True

	def __str__(self):
		"""
		<par>Return a nice string representation of <self/>. This is used
		in the progress report.</par>
		"""
		try:
			name = self.__class__.name
		except AttributeError:
			name = self.__class__.__name__
		return "%s(%s)" % (name, self.project.idStr(self.id))

	def __repr__(self):
		return "<%s object id=%r at 0x%x>" % (self.__class__.__name__, self.id, id(self))

	def failed(self):
		"""
		Called by actions when building a target fails when writing the output
		has already begun. Derived classes will overwrite <method>failed</method>
		to remove broken files.
		"""
		self.project.report("Deleting broken target %s ...\n" % self)


class FileSystemTarget(Target):
	"""
	<par>A target in the filesystem, i.e. a file or directory.</par>
	"""
	pass


class FileTarget(FileSystemTarget):
	"""
	<par>a file</par>
	"""
	name = "File"

	def _fetchlastmodified(self):
		try:
			return self.id.mtime()
		except OSError:
			return None

	def failed(self):
		super(FileTarget, self).failed()
		self.id.remove()


class DirectoryTarget(FileSystemTarget):
	"""
	<par>a directory.</par>
	"""
	name = "Dir"

	def _fetchlastmodified(self):
		# only rebuild if it doesn't exist, if it exists do nothing
		try:
			timestamp = self.id.mtime()
		except OSError:
			return None
		else:
			return datetime.datetime(1900, 1, 1) # should be older than any existing file/directory

	def hasToBeRebuilt(self):
		return self.lastmodified() is None


class ImageTarget(FileTarget):
	"""
	<par>an image file.</par>
	"""
	name = "Image"


class JavascriptTarget(FileTarget):
	"""
	<par>a Javascript source file (extension <lit>.js</lit>)</par>
	"""
	name = "JS"


class CascadingStyleSheetTarget(FileTarget):
	"""
	<par>a Cascading Stylesheet file (extension <lit>.css</lit>)</par>
	"""
	name = "CSS"


class PHPTarget(FileTarget):
	"""
	<par>a PHP include file (extension <lit>.php</lit>)</par>
	"""
	name = "PHP"


class HTMLTarget(FileTarget):
	"""
	<par>a &html; file (extension <lit>.html</lit>, <lit>.phtml</lit>, <lit>.jsp</lit>, etc.)</par>
	"""
	name = "HTML"

	def __init__(self, project, id, actions=None, base=None):
		FileTarget.__init__(self, project, id, actions)
		self.base = base


class XMLTarget(FileTarget):
	name = "XML"

	def __init__(self, project, id, actions=None, base=None):
		FileTarget.__init__(self, project, id, actions)
		self.base = base


class DTDTarget(FileTarget):
	"""
	<par>a Document Type Definition file (extension <lit>.dtd</lit>)</par>
	"""
	name = "DTD"


class PDFTarget(FileTarget):
	"""
	<par>a &pdf; file (extension <lit>.pdf</lit>)</par>
	"""
	name = "PDF"


class JavaPropTarget(FileTarget):
	"""
	<par>a Java properties file (extension <lit>.properties</lit>)</par>
	"""
	name = "JavaProp"


class SQLTarget(FileTarget):
	"""
	<par>an &sql; script (extension <lit>.sql</lit>)</par>
	"""
	name = "SQL"


class PythonTarget(FileTarget):
	"""
	<par>a Python source file (extension <lit>.py</lit>)</par>
	"""
	name = "Python"

	def __init__(self, project, id, actions=None):
		FileTarget.__init__(self, project, id, actions)
		self.module = None

	def dirty(self):
		FileTarget.dirty(self)
		if self.module is not None:
			url.removefromimportcache(self.module)
			self.module = None

	def load(self, mode="new"):
		"""
		<par>import and return the Python module as the result.</par>
		"""
		if self.module is None:
			self.module = self.id.import_(mode)
		return self.module


class XISTNSTarget(PythonTarget):
	"""
	<par>a Python source file that contains a &xist; <pyref module="xist.xsc" class="Namespace"><class>Namespace</class></pyref>
	object and must be imported to provided element definitions etc. to the &xist; parser.</par>
	"""
	name = "XISTNS"

	def __init__(self, project, id, actions=None, prefix=None):
		PythonTarget.__init__(self, project, id, actions)
		self.prefix = prefix

# backwards compatibility
XISTNamespaceTarget = XISTNSTarget


class XISTTarget(FileTarget):
	"""
	<par>an &xist; source file, i.e. a &xml; file that has to be transformed via <pyref module="ll.xist">&xist;</pyref>
	(extensions <lit>.xmlxsc</lit>, <lit>.htmlxsc</lit>, etc.)</par>
	"""
	name = "XIST"

	def __init__(self, project, id, actions=None, base=None, parser=None, handler=None, prefixes=None):
		FileTarget.__init__(self, project, id, actions)
		self.base = base
		self.parser = parser
		self.handler = handler
		self.prefixes = prefixes

	def parse(self):
		from ll.xist import parsers
		# FIXME This is for backwards compatibility with XIST 1.5
		if self.prefixes is not None:
			return parsers.parseURL(self.id, base=self.base, parser=self.parser, handler=self.handler, prefixes=self.prefixes)
		else:
			return parsers.parseURL(self.id, base=self.base, parser=self.parser, handler=self.handler)


class PhonyTarget(Target):
	"""
	<par>A <class>PhonyTarget</class> does not really exist. It can
	be used as a virtual target for all the real targets that have to be built.</par>
	"""
	name = "Phony"

	def __init__(self, project, id, actions=None, doc=None):
		"""
		<par>create a new <class>PhonyTarget</class>.</par>

		<par>If <arg>actions</arg> is not given or <lit>None</lit> the
		action for rebuilding <self/> will be a <pyref class="NoAction"><class>NoAction</class></pyref>
		instance.</par>

		<par><arg>doc</arg> should be a description what this <class>PhonyTarget</class> and will
		be printed when <pyref class="Project" method="buildWithArgs"><method>buildWithArgs</method></pyref>
		is called without arguments.</par>
		"""
		if actions is None:
			actions = NoAction()
		Target.__init__(self, project, id, actions)
		self.doc = doc

	def _fetchlastmodified(self):
		"""
		<par>get the <z>last modification</z> timestamp. This will always
		be <lit>None</lit>, as a <class>PhonyTarget</class>
		does not really exist. Note that this means that the associated action will be executed
		every time the target is updated.</par>
		"""
		return None


###
### Action classes
###

class Action(object):
	"""
	<par>An <class>Action</class> is responsible for rebuilding or modifying a
	<pyref class="Target"><class>Target</class></pyref>.</par>

	<par>An <class>Action</class> is stateless and so may be used by multiple targets. Executing the
	action is done through a call to the method <pyref method="execute"><method>execute</method></pyref>.</par>

	<par>There are two types of actions:</par>
	<ulist>
	<item>Primary actions: Those actions create the target object.</item>
	<item>Secondary actions: Those actions modify a target created by a primary
	action. The subclasses <pyref class="ModeAction"><class>ModeAction</class></pyref>
	and <pyref class="OwnerAction"><class>OwnerAction</class></pyref> are examples
	of secondary actions.</item>
	</ulist>
	"""
	def __init__(self):
		"""
		<par>create a new <class>Action</class> instance.</par>
		"""
		pass

	def execute(self, target):
		"""
		<par>execute the action to rebuild or modify <arg>target</arg> and
		return a one-line status/result message (or <lit>None</lit> if there
		is nothing special to report).</par>
		"""
		raise NotImplementedError()

	def __str__(self):
		"""
		<par>Return a string description for <self/>. This is the class name unless
		there is a class attribute <lit>name</lit>. This string is used in the progress report.</par>
		"""
		try:
			return self.__class__.name
		except AttributeError:
			return self.__class__.__name__


class CopyAction(Action):
	"""
	<par>generate the target by copying the source.</par>
	<par>The <z>source</z> is the only <pyref class="Target"><class>Target</class></pyref> on which
	the target depends through a <pyref class="MainDep"><class>MainDep</class></pyref>.</par>
	"""
	name = "Copy"

	def execute(self, target):
		"""
		<par>execute the action</par>
		"""
		sources = target.sources(MainDep)
		assert len(sources)==1, "copy must have one main source for target %r, has %r" % (target, sources)

		infile = sources[0].id.openread()
		data = infile.read()
		infile.close()
		try:
			outfile = None
			outfile = target.id.openwrite()
			outfile.write(data)
		except:
			if outfile is not None:
				outfile.close()
			target.failed()
			raise
		else:
			if outfile is not None:
				outfile.close()
		return "%s copied" % target.project.sizeStr(len(data))


class JoinAction(Action):
	"""
	<par>generate the target by joining the sources.</par>
	<par>The <z>sources</z> must depend on the target via a
	<pyref class="OrderedDep"><class>OrderedDep</class></pyref> and will be join
	in the order specified by these dependencies.</par>
	"""
	name = "Join"

	def execute(self, target):
		"""
		<par>execute the action</par>
		"""
		deps = target.sourcedeps(OrderedDep)
		deps.sort(lambda s1, s2: cmp(s1.order, s2.order))

		try:
			outfile = None
			outfile = target.id.openwrite()
			size = 0
			for dep in deps:
				infile = dep.source.id.openread()
				data = infile.read()
				infile.close()
				outfile.write(data)
				size += len(data)
		except:
			if outfile is not None:
				outfile.close()
			target.failed()
			raise
		else:
			if outfile is not None:
				outfile.close()
		return "%s joined" % target.project.sizeStr(size)


class MkDirAction(Action):
	"""
	<par>This action creates the target as a directory.</par>
	"""
	name = "MkDir"

	def __init__(self, mode=0777):
		"""
		Create a <class>MkDirAction</class> instance. <arg>mode</arg> (which defaults
		to <lit>0777</lit>) will be used as the permission bit pattern for the new directory.
		"""
		super(MkdDirAction, self).__init__()
		self.mode = mode

	def execute(self, target):
		"""
		<par>create the directory with the permission bits specified in the constructor.</par>
		"""
		target.id.makedirs(self.mode)

	def __repr__(self):
		return "<%s.%s mode=%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, oct(self.mode), id(self))

	def __str__(self):
		return "%s(mode=%s)" % (super(MkDirAction, self).__str__(), oct(self.mode))


class XISTAction(Action):
	"""
	<par>This action creates the target via transforming the main source
	via <pyref module="ll.xist"><app>&xist;</app></pyref>.</par>
	"""
	name = "XIST"

	def __init__(self, encoding=None, xhtml=None, mode=None, target=None, stage=None, lang=None, targetroot=None):
		"""
		<par>create a new <class>XISTAction</class> instance. The arguments
		will be used during parsing, transformation and publishing in the usual way.</par>
		"""
		super(XISTAction, self).__init__()
		self.encoding = encoding
		self.xhtml = xhtml
		self.mode = mode
		self.target = target
		self.stage = stage
		self.lang = lang
		self.targetroot = targetroot

	def loadNamespaces(self, target):
		"""
		<par>import all the source namespaces (i.e. all the sources of <self/> that are instances
		of <pyref class="XISTNamespaceTarget"><class>XISTNamespaceTarget</class></pyref>).</par>
		<par>The module objects will be returned in a list.</par>
		"""
		namespaces = [ source.load() for source in target.sources() if isinstance(source, XISTNamespaceTarget) ]
		return namespaces

	def converter(self, maketarget=None):
		"""
		<par>create a new <pyref module="ll.xist.converters" class="Converter"><class>Converter</class></pyref> instance
		for this action. The attributes of this new converter (<lit>mode</lit>, <lit>target</lit>,
		<lit>stage</lit>, etc.) will correspond to those specified in the constructor.</par>
		<par>When the argument <arg>target</arg> is specified it will be used to initialize the
		attribute <lit>maketarget</lit> of the converter.</par>
		"""
		from ll.xist import converters
		return converters.Converter(root=self.targetroot, mode=self.mode, stage=self.stage, target=self.target, lang=self.lang, makeaction=self, maketarget=maketarget)

	def convert(self, node, maketarget=None):
		return node.convert(self.converter(maketarget))

	def publisher(self, target, stream):
		"""
		<par>create a <pyref module="ll.xist.publishers" class="Publisher"><class>Publisher</class></pyref>
		that will be used for generating the target <arg>target</arg>. The publisher must use
		<arg>stream</arg> as it's output stream.</par>
		"""
		from ll.xist import publishers
		return publishers.FilePublisher(stream, base=target.base, root=self.targetroot, encoding=self.encoding, xhtml=self.xhtml)

	def publish(self, target, node):
		"""
		<par>publish the <arg>node</arg> as the content of the
		target <arg>target</arg>.</par>
		"""
		outfile = target.id.openwrite()
		from ll.xist import publishers
		publisher = self.publisher(target, outfile)
		try:
			publisher.doPublication(node)
		except:
			target.failed()
			raise
		bytes = publisher.tell()
		outfile.close()
		return bytes

	def execute(self, target):
		"""
		<par>execute the action. The <arg>target</arg> must have exactly one source
		with a <pyref class="MainDep"><class>MainDep</class></pyref> dependency. This target will be transformed
		using all the namespaces available via <pyref method="loadNamespaces"><method>loadNamespaces</method></pyref>
		and the resulting node will be published via <pyref method="publish"><method>publish</method></pyref>.</par>
		"""
		self.loadNamespaces(target)
		xistsources = target.sources(MainDep)
		assert len(xistsources)==1, "XISTAction must have one main source for target %r, has %r" % (target, xistsources)
		t1 = dtnow()
		node = xistsources[0].parse()
		t2 = dtnow()
		node = self.convert(node, maketarget=target)
		t3 = dtnow()
		bytes = self.publish(target, node)
		t4 = dtnow()
		return "parse %s; convert %s; save %s; %s published" % (target.project.secondsStr(t2-t1), target.project.secondsStr(t3-t2), target.project.secondsStr(t4-t3), target.project.sizeStr(bytes))

	def __repr__(self):
		options = []
		for optionname in ("encoding", "xhtml", "mode", "target", "stage", "lang", "targetroot"):
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append(" %s=%r" % (optionname, optionvalue))
		return "<%s.%s%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, "".join(options), id(self))

	def __str__(self):
		options = []
		for optionname in ("encoding", "xhtml", "mode", "target", "stage", "lang", "targetroot"):
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append("%s=%r" % (optionname, optionvalue))
		return "%s(%s)" % (super(XISTAction, self).__str__(), ", ".join(options))


class SplatAction(Action):
	"""
	<par>This action creates the target via replacing certain regular expressions
	in the main source.</par>
	"""
	name = "Splat"

	def __init__(self, patterns):
		"""
		<par>create a new <class>SplatAction</class> instance. <arg>patterns</arg>
		is a list of pattern pairs. The first entry will be replaced by the second entry.</par>
		"""
		super(SplatAction, self).__init__()
		self.patterns = patterns

	def execute(self, target):
		"""
		<par>execute the action. The <arg>target</arg> must have exactly one source
		with a <pyref class="MainDep"><class>MainDep</class></pyref> dependency.</par>
		"""
		sources = target.sources(MainDep)
		assert len(sources)==1, "SplatAction must have one main source for target %r, has %r" % (target, sources)
		t1 = dtnow()
		data = sources[0].id.openread().read()
		t2 = dtnow()
		for (search, replace) in self.patterns:
			data = re.sub(search, replace, data)
		t3 = dtnow()
		file = target.id.openwrite()
		try:
			file.write(data)
		except:
			target.id.remove()
			raise
		bytes = file.tell()
		file.close()
		t4 = dtnow()
		return "read %s; replace %s; save %s; %s written" % (target.project.secondsStr(t2-t1), target.project.secondsStr(t3-t2), target.project.secondsStr(t4-t3), target.project.sizeStr(bytes))

	def __repr__(self):
		return "<%s.%s patterns=%r at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, self.patterns, id(self))

	def __str__(self):
		return "Splat(%r)" % (self.patterns,)


class NoAction(Action):
	"""
	<par>This action does nothing. It can be used as the action for <pyref class="PhonyTarget"><class>PhonyTarget</class>s</pyref>.</par>
	"""
	name = "None"

	def execute(self, target):
		"""
		<par>do nothing</par>
		"""
		pass


class CommandAction(Action):
	"""
	<par>This action executes a system command (via <function>os.system</function>).</par>
	"""
	name = "Command"

	def __init__(self, command):
		"""
		<par>create a new <class>CommandAction</class>. <arg>command</arg> is the command
		that will executed when <pyref method="execute"><method>execute</method></pyref> is called.</par>
		"""
		super(CommandAction, self).__init__()
		self.command = command

	def execute(self, target):
		"""
		<par>execute the action (note that the <arg>target</arg> is irrelavant and will be ignored;
		in most cases it will probably be a <pyref class="PhonyTarget"><class>PhonyTarget</class></pyref>).</par>
		"""
		os.system(self.command)

	def __repr__(self):
		return "<%s.%s command=%r at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, self.command, id(self))

	def __str__(self):
		return "%s(%r)" % (super(CommandAction, self).__str__(), self.command)


class ModeAction(Action):
	"""
	<class>ModeAction</class> is used for changing file permission.
	"""
	name = "Mode"

	def __init__(self, mode=0644):
		"""
		Create an <class>ModeAction</class> instance. <arg>mode</arg> (which defaults to <lit>0644</lit>) will
		be use as the permission bit pattern.
		"""
		super(ModeAction, self).__init__()
		self.mode = mode

	def execute(self, target):
		target.id.chmod(self.mode)

	def __repr__(self):
		return "<%s.%s mode=%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, oct(self.mode), id(self))

	def __str__(self):
		return "%s(mode=%s)" % (super(ModeAction, self).__str__(), oct(self.mode))


class OwnerAction(Action):
	"""
	<class>OwnerAction</class> is used for changing the the user and/or group ownership of files.
	"""
	name = "Owner"

	def __init__(self, user=None, group=None):
		"""
		Create a new <class>OwnerAction</class> instance. <arg>user</arg> can either be a numerical
		user id or a user name or <lit>None</lit>. If it is <lit>None</lit> no user ownership will
		be changed. The same applies to <arg>group</arg>.
		"""
		super(OwnerAction, self).__init__()
		self.user = user
		self.group = group

	def execute(self, target):
		target.id.chown(self.user, self.group)

	def __repr__(self):
		v = []
		if self.user is not None:
			v.append(" user=%r" % self.user)
		if self.group is not None:
			v.append(" group=%r" % self.group)
		return "<%s.%s%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, "".join(v), id(self))

	def __str__(self):
		v = []
		if self.user is not None:
			v.append("user=%r" % self.user)
		if self.group is not None:
			v.append("group=%r" % self.group)
		return "%s(%s)" % (super(OwnerAction, self).__str__(), ", ".join(v))


###
### Classes for target ids (apart from strings for PhonyTargets and URLs for file targets)
###

class DBID(object):
	"""
	<par>This class provides a unique identifier for database content. This
	can be used as an id for <pyref class="Target"><class>Target</class></pyref>
	objects that are not files, but database records.</par>
	"""
	def __init__(self, type, host=None, port=None, user=None, database=None, table=None, key=None):
		self.type = type
		self.host = host
		self.port = port
		self.user = user
		self.database = database
		self.table = table
		self.key = key

	def __eq__(self, other):
		res = self.__class__ == other.__class__
		if not res:
			res = self.type==other.type and self.host==other.host and self.port==other.port and self.user==other.user and self.database==other.database and self.table==other.table and self.key==other.key
		return res

	def __hash__(self):
		return hash(self.type) ^ hash(self.host) ^ hash(self.port) ^ hash(self.user) ^ hash(self.database) ^ hash(self.table) ^ hash(self.key)

	def __repr__(self):
		args = []
		for attrname in ("type", "host", "port", "user", "database", "table", "key"):
			attrvalue = getattr(self, attrname)
			if attrvalue is not None:
				args.append("%s=%r" % (attrname, attrvalue))
		return "%s(%s)" % (self.__class__.__name__, ", ".join(args))


###
### The project class
###

class Project(dict):
	"""
	<par>A <class>Project</class> collects all <pyref class="Target"><class>Target</class></pyref>s
	from a project. It is responsible for initiating the build process and for generating a report about
	the progress of the build process.</par>
	"""
	def __init__(self, showSummary=1, showTargets=1, showActions=0, showSources=0, showMessages=0, ignoreExceptions=0, color=None, maxNewerReport=2):
		super(Project, self).__init__()
		self.targetsrebuilt = 0
		self.targetsfailed = 0
		self.showSummary = showSummary
		self.showTargets = showTargets
		self.showActions = showActions
		self.showSources = showSources
		self.showMessages = showMessages
		self.ignoreExceptions = ignoreExceptions
		if color is None:
			color = os.getenv("MAKE_REPRANSI")
			if color is None:
				color = 0
			else:
				color = int(color)
				if not sys.platform.startswith("linux") and not sys.platform.startswith("darwin"):
					color = 0
		self.color = color
		self.maxNewerReport = maxNewerReport
		self._initColors()
		self.__built = False # remember whether build() has been called, so we can reset timestamps on the second call

	def __repr__(self):
		return "<%s.%s with %d targets at 0x%x>" % (self.__module__, self.__class__.__name__, len(self), id(self))

	def _initColors(self):
		"""
		<par>initialize the color classes which are used for the various formatting
		methods.</par>
		"""
		self.TextTime = ansistyle.Text5
		self.TextSize = ansistyle.Text4
		self.TextID = ansistyle.Text2
		self.TextAction = ansistyle.Text15
		self.TextError = ansistyle.Text9

	def secondsStr(self, delta):
		"""
		<par>return a nicely formatted and colored string for
		the <class>datetime.timedelta</class> value <arg>delta</arg>.</par>
		"""
		seconds = delta.seconds+1e-7*delta.microseconds

		if self.color:
			return ansistyle.Text(self.TextTime("%.03f" % seconds, "s"))
		else:
			return "%.03fs" % seconds

	def sizeStr(self, bytes):
		"""
		<par>return a nicely formatted and colored string for
		the byte count value <arg>bytes</arg>.</par>
		"""
		if self.color:
			return ansistyle.Text(self.TextSize("%d" % bytes), " bytes")
		else:
			return "%d bytes" % bytes

	def actionStr(self, action):
		"""
		<par>return a nicely formatted and colored string for
		the action <arg>action</arg>.</par>
		"""
		s = str(action)
		if self.color:
			s = self.TextAction(s)
		return s

	def errorStr(self, text):
		"""
		<par>return a nicely formatted and colored string for
		the error text <arg>text</arg>.</par>
		"""
		if self.color:
			return self.TextError(text)
		else:
			return text

	def idStr(self, id):
		"""
		<par>return a nicely formatted and colored string for
		the id <arg>id</arg>.</par>
		"""
		s = str(id)
		if self.color:
			s = self.TextID(s)
		return s

	def __setitem__(self, id, target):
		"""
		<par>set the target with the id <arg>name</arg>.</par>
		"""
		if id in self:
			self.reportWarning(RedefinedTargetWarning(id), 5)
		if isinstance(id, url.URL):
			id = id.abs(scheme="file")
		super(Project, self).__setitem__(id, target)

	def __getitem__(self, id):
		"""
		<par>return the target with the id <arg>id</arg>.</par>
		<par>If an id can't be found, it will be wrapped in a <pyref module="ll.url" class="URL"><class>URL</class></pyref> instance and retried.</par>
		"""
		sup = super(Project, self)
		try:
			return sup.__getitem__(id)
		except KeyError:
			if isinstance(id, (str, unicode)):
				id = url.URL(id).abs(scheme="file")
			elif isinstance(id, url.URL):
				id = id.abs(scheme="file")
			else:
				raise UndefinedTargetError(id)
			try:
				return sup.__getitem__(id) # retry with a filename
			except KeyError:
				id = id.real(scheme="file")
				try:
					return sup.__getitem__(id) # retry with real filename
				except KeyError:
					raise UndefinedTargetError(id)

	def has_key(self, id):
		"""
		<par>Return whether the target with the id <arg>id</arg>
		exists in the project.</par>
		"""
		return id in self

	def __contains__(self, id):
		"""
		<par>Return whether the target with the id <arg>id</arg>
		exists in the project.</par>
		"""
		sup = super(Project, self)
		has = sup.has_key(id)
		if not has:
			if isinstance(id, (str, unicode)):
				id = url.URL(id).abs(scheme="file")
			elif isinstance(id, url.URL):
				id = id.abs(scheme="file")
			has = sup.has_key(id) # retry with a filename
			if not has:
				id = id.real(scheme="file")
				has = sup.has_key(id) # retry with real filename
		return has

	def has(self, id):
		warnings.warn("has() is deprecated, use has_key() or __contains__() (i.e. ... in ...) instead", DeprecationWarning)
		return self.has_key(id)

	def clear(self):
		"""
		<par>Removes as targets from <self/>. If you keep external references
		to targets, they will be worthless, because the dependencies between
		target will be removed.</par>
		"""
		for target in self.itervalues():
			target.clear()
		super(Project, self).clear()

	def create(self):
		"""
		<par>Create all dependencies for the project. Overwrite in subclasses.</par>
		<par>This method should only be called once, otherwise you'll get lots of
		<pyref class="RedefinedTargetWarning"><class>RedefinedTargetWarning</class>s</pyref>.
		But you can call <pyref method="clear"><method>clear</method></pyref> to
		remove all targets before calling <method>create</method>. You can also
		use the method <method>recreate</method> for that.</par>
		"""

	def recreate(self):
		"""
		<par>Calls <pyref method="clear"><method>clear</method></pyref> and
		<pyref method="create"><method>create</method></pyref>.</par>
		"""
		self.clear()
		self.create()

	def build(self, *targets):
		"""
		<par>build all targets in <arg>targets</arg>.</par>
		
		<par>Items in <arg>targets</arg> must be <pyref class="Target"><class>Target</class></pyref>
		instances or target ids.</par>

		<par>This method can be called multiple times and will recheck modification timestamp
		on each call.</par>
		"""
		t1 = dtnow()
		self.targetsrebuilt = 0
		self.targetsfailed = 0
		if self.__built:
			for target in self.itervalues():
				target.dirty()
		self.__built = True # remember the call here, otherwise we'd get half initialized timestamps in case of an error, but won't clear them on the second call to build()
		for target in targets:
			if not isinstance(target, Target):
				target = self[target]
			target.update()
		t2 = dtnow()
		if self.showSummary:
			self.report("build %s\n%s%s (%d targets; %d rebuilt; %d failed)\n" % (self.__class__.__name__, indent, self.secondsStr(t2-t1), len(self), self.targetsrebuilt, self.targetsfailed))

	def buildWithArgs(self, args):
		"""
		For calling make scripts from the command line. <arg>args</arg> can be <lit>sys.argv[1:]</lit> or something equivalent.
		"""
		if args:
			# FIXME: Use optparse
			(options, args) = getopt.getopt(args, "t:s:a:m:z:ic", ["showtargets=", "showsources=", "showactions=", "showmessages=", "showsummary=", "ignore", "color"])

			for (option, value) in options:
				if option=="-t" or option=="--showtargets":
					self.showTargets = int(value)
				if option=="-s" or option=="--showsources":
					self.showSources = int(value)
				if option=="-a" or option=="--showactions":
					self.showActions = int(value)
				if option=="-m" or option=="--showmessages":
					self.showMessages = int(value)
				if option=="-z" or option=="--showsummary":
					self.showSummary = int(value)
				if option=="-i" or option=="--ignore":
					self.ignoreExceptions = 1
				if option=="-c" or option=="--color":
					self.color = 1
			self.build(*args)
		else:
			self.report("Available phony targets are:\n")
			phonies = []
			maxlen = 0
			for id in self:
				if isinstance(id, basestring):
					maxlen = max(maxlen, len(id))
					phonies.append(self[id])
			phonies.sort(lambda p1, p2: cmp(p1.id, p2.id)) # FIXME: Use sort(key=...) in Python 2.4
			for phony in phonies:
				if phony.doc:
					id = ("%s %s" % (phony.id, "."*(maxlen+4)))[:maxlen+3] + " " # FIXME: Use str.rjust(fill=...) in Python 2.4
					doc = phony.doc
				else:
					id = phony.id
					doc = ""
				self.report("\t%s%s\n" % (id, doc))

	def report(self, text):
		"""
		All screen output is done through this methods. This makes it possible to redirect the output
		(e.g. to logfiles) in subclasses.
		"""
		sys.stderr.write(text)

	def reportError(self, text):
		"""
		Report an error
		"""
		self.report(text)

	def reportWarning(self, warning, stacklevel):
		"""
		Issue a warning through the Python warnings framework
		"""
		warnings.warn(warning, stacklevel=stacklevel)
